/* Name: main.c
 * Author: Jeff Keyzer	http://mightyohm.com 
 * Date: 01/30/09
 * License: Creative Commons Attribution-Share Alike 3.0
 *
 * -- User Interface for Wifi Radio Project --
 * This program accepts serial data from a shell script running on an ASUS WL-520GU wireless
 * router.  The shell script queries a local mpd server and returns the name of the current
 * radio stream as well as the artist/title of the current song.  The AVR expects to get
 * two lines of serial data with each query, one for the stream name prefixed with "Name: "
 * and one for the artist/title prefixed "Title: " (in that order)
 *
 * A TIMER1 overflow interrupt routine periodically checks the position of a tuner
 * potentiometer attached to an analog input of the AVR.  If the tuner position has
 * changed by more than a preset # of ADC counts, the AVR sends the new tuner value
 * back to the router, where a script changes the currently playing station.
 *
 * Built and tested with an ATmega168 @ 16MHz and a Sparkfun 16x2 character LCD display
 *
 * More information is available at
 * http://mightyohm.com/blog/
 */

// Includes
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <string.h>
#include <stdio.h>

#include "lcd.h"				// Peter Fleury's LCD Library
								// http://homepage.hispeed.ch/peterfleury/avr-lcd44780.html

// Defines
#define	BAUD			9600	// USART baud rate, must agree with router ttyS0 settings
#define	SER_BUFF_LEN	80		// longest character line to accept from serial port
#define	LCD_WIDTH		20		// visible width of LCD display
#define DDRAM_WIDTH		40		// internal line length of LCD display
#define	PAGEDELAY		3000	// delay between LCD pages, in ms
#define SHIFTDELAY		300		// horizontal shift rate of LCD when display > visible area

#define	ADC_SENS		5		// # of counts ADC value can move before sending position

// Function prototypes
void inituart(void);
void uart_putchar(unsigned char data);
void uart_putchar(unsigned char data);
unsigned char uart_getchar(void);
void uart_flush(void);
void getline(unsigned char *serbuffer);
void putstring(unsigned char *buffer);
void lcd_putvis(unsigned char *s);
void init_adc(void);
int read_adc(void);
void init_timer1(void);
void timer1_overflow(void);

char* lineprint(char *text, char* buffer, int position);
char* StringPadRight(char *string, int padded_len, char *pad);
char* StringPadLeft(char *string, int padded_len, char *pad);

// Global variables
unsigned char serbuffer[SER_BUFF_LEN];	// serial buffer
int tunerpos;	// tuner knob position
int oldtunerpos = -99;	// last position
int abort_scroll = 0;	// ISR sets this if station is changed, aborts LCD scrolling routine

// Timer1 overflow interrupt service routine (ISR)
SIGNAL (TIMER1_OVF_vect) // SIGNAL call makes sure we don't interrupt the interrupt
{
	TCNT1 = 0x10000 - (F_CPU/1024/2);	// reset timer
	tunerpos = read_adc();	// get ADC value corresponding to tuner position
	
	// if the knob moves, send new position over serial port
    /*(
   if ( tunerpos == 1023 ) {
	//if (tunerpos > oldtunerpos+ADC_SENS || tunerpos < oldtunerpos-ADC_SENS) {	
		sprintf(serbuffer, "tuner: %d\n", tunerpos);	// convert ADC value to string
		putstring(serbuffer);	// transmit over serial link
      abort_scroll = 1;		// tell LCD routine to stop scrolling and show new station name
	}
      */
   
}

// Start of main program
int main(void)
{
	
    unsigned char artist[SER_BUFF_LEN];		// name of current radio station
    unsigned char track[SER_BUFF_LEN];		// name of current radio station
    unsigned char station[SER_BUFF_LEN];		// name of current radio station
    unsigned char state[SER_BUFF_LEN];		// artist & title of current song
    unsigned char temp[SER_BUFF_LEN];		// artist & title of current song
    unsigned char buffer[160];		// artist & title of current song
    int maxlen = 0;
    int i, extra;
	
	/* initialization routines */
    inituart();				// initialize AVR serial port (USART0)
    init_adc();
    
    // for some reason the 1st character after reset gets mangled, so send out a
    // newline to flush things out
    uart_putchar('\n');
 	putstring("AVR Online\n");	// helps with debugging to see if AVR is alive
    
    // initialize LCD display
    lcd_init(LCD_DISP_ON);
    lcd_clrscr();
    
    // Display splash screen
    lcd_puts("       cub\n        labs\n \n     wifi radio");
    _delay_ms(PAGEDELAY);
    lcd_clrscr();
    lcd_puts(" \n .. starting up ..\n   hold puhleeze\n \n");
    _delay_ms(PAGEDELAY);
    
    // wait for router to tell us to start (avoids flooding the router with data
    // before it's ready)
    do getline(serbuffer);
    while (!strstr(serbuffer, "AVR Start!"));
    
    //lcd_puts("Go!\n");
    
    // Launch tuner position read interrupt routine
    init_timer1();
    sei();		// enable interrupts
    

    /* main program loop */
    for(;;){	// loop forever
		
		abort_scroll = 0;	// reset
		
		// Grab name and title data from the router
		// Note: program execution will stall until serial data is received!

		do getline(serbuffer);	
		while (!strstr(serbuffer, "Artist: "));
		strncpy(artist, serbuffer+(sizeof("Artist: ")-1),DDRAM_WIDTH);
		do getline(serbuffer);	
		while (!strstr(serbuffer, "Track: "));
		strncpy(track, serbuffer+(sizeof("Track: ")-1),DDRAM_WIDTH);
		do getline(serbuffer);	
		while (!strstr(serbuffer, "Station: "));
		strncpy(station, serbuffer+(sizeof("Station: ")-1),DDRAM_WIDTH);
		do getline(serbuffer);	
		while (!strstr(serbuffer, "State: "));
		strncpy(temp, serbuffer+(sizeof("State: ")-1),DDRAM_WIDTH);


      strcpy(state,StringPadLeft(temp,LCD_WIDTH,' '));

      if (strlen(artist) > maxlen) maxlen = strlen(artist);
      if (strlen(track) > maxlen) maxlen = strlen(track);
      if (strlen(state) > maxlen) maxlen = strlen(state);
      if (strlen(station) > maxlen) maxlen = strlen(station);

      char str[25] = {'\0'} ;

      if ( maxlen > LCD_WIDTH ) {
         extra = maxlen - LCD_WIDTH;
         for (i = 0; i <= extra; i++) {
            if (abort_scroll == 1) break;
            strcpy(buffer,lineprint(artist,str,i));
            strcat(buffer,lineprint(state,str,i));
            strcat(buffer,lineprint(track,str,i));
            strcat(buffer,lineprint(station,str,i));
            lcd_clrscr();
            lcd_home();
            lcd_puts(buffer);
            _delay_ms(SHIFTDELAY);
         }   
         for (i = extra-1; i >= 0; i--) {
            if (abort_scroll == 1) break;
            strcpy(buffer,lineprint(artist,str,i));
            strcat(buffer,lineprint(state,str,i));
            strcat(buffer,lineprint(track,str,i));
            strcat(buffer,lineprint(station,str,i));
            lcd_clrscr();
            lcd_home();
            lcd_puts(buffer);
            _delay_ms(SHIFTDELAY);
         }
      } else {
            strcpy(buffer,lineprint(artist,str,i));
            strcat(buffer,lineprint(state,str,i));
            strcat(buffer,lineprint(track,str,i));
            strcat(buffer,lineprint(station,str,i));
            lcd_clrscr();
            lcd_home();
            lcd_puts(buffer);
            _delay_ms(SHIFTDELAY);
      }
   }
   return 0;
}

// Functions
void inituart(void)	// Initialize USART0 to desired baud rate
{
	// Set baud rate generator based on F_CPU
	UBRR0H = (unsigned char)(F_CPU/(16UL*BAUD)-1)>>8;
	UBRR0L = (unsigned char)(F_CPU/(16UL*BAUD)-1);
	
	// Enable USART0 transmitter and receiver
	UCSR0B = (1<<RXEN0) | (1<<TXEN0);
	
}

unsigned char uart_getchar(void)	// get a character from the serial port
{

	while(!(UCSR0A & (1 << RXC0)))	// wait until character is received
		;
	return UDR0;
}

void uart_putchar(unsigned char data)	// send a character over the serial port
{

	while(!(UCSR0A & (1 << UDRE0))) // wait until UART is ready for tx
		;       // do nothing
	UDR0 = data;    // send char

}

void uart_flush(void)	// flush RX buffer
{
	unsigned char dummy;
	
	while(UCSR0A & (1<<RXC0))	// while there is a character in the RX queue
		dummy = UDR0;			// flush the queue by reading it
}

void getline(unsigned char *buffer)	// get a single line from the UART rcvr
{
	int bufpos = 0;
	
	uart_flush();	// clear RX buffer of stale chars if present
	//cli();
	// read a line from UART, stop if buffer is full or we reach newline
	while ((bufpos < (SER_BUFF_LEN - 1)) && (buffer[bufpos] = uart_getchar()) != '\n') {

		if (buffer[bufpos] == '\r')
			bufpos--;	// strip carriage returns in case data contains both CR&LF
			
		bufpos++;
	}
	
	buffer[bufpos] = '\0';	// turn buffer into a string
	//sei();
}

void putstring(unsigned char *buffer)	// send a string to the UART
{
	int bufpos = 0;
	
	cli();	// disable interrupts so we don't mangle the data
	//uart_flush();	// clear RX buffer of stale chars if present
	
	// start sending characters over the serial port until we reach the end of the string
	while ((bufpos < (SER_BUFF_LEN - 1)) && (buffer[bufpos] != '\0')) {

		uart_putchar(buffer[bufpos]);
			
		bufpos++;
	}
	
	//uart_putchar('\n');
	sei();	// enable interrupts again
	
}

void init_adc(void) {	// configure ADC

	//PRR &= ~(1 << PRADC);	

	ADCSRA |= (1 << ADEN) | 0x07;	// enable ADC, max prescaler
	ADMUX |= (1 << REFS0) | (1 << MUX2);	// Select ADC4, internal reference
}

int read_adc(void) {	// read 2 bytes (only 10 useful bits) from selected ADC channel
	uint8_t low, high;

	// start a conversion and wait for it to finish
	ADCSRA |= (1 << ADSC);
	while(ADCSRA & (1 << ADSC));
	
	// store the results
	low = ADCL;
	high = ADCH;
	
	return (high << 8) | low;	// return 16 bit value
}

void init_timer1(void) {
	// initialize TIMER1 to trigger an overflow interrupt every 0.5sec
	TCCR1A = 0;
	TCCR1B |= (1 << CS12) | (1 << CS10);

	
	TCNT1 = 0x10000 - (F_CPU/1024/2);
	
	TIFR1=0;
	TIMSK1 |= (1 << TOIE1);
}

char *StringPadRight(char *string, int padded_len, char *pad) {
   int len = (int) strlen(string);
   if (len >= padded_len) {
      return string;
   }   
   int i;
   for (i = 0; i < padded_len - len; i++) {
      strcat(string, pad);
   }   
   return string;
}

char *StringPadLeft(char *string, int padded_len, char *pad) {
   int len = (int) strlen(string);
   int i;
   if (strlen(string) < padded_len) strcpy(string,StringPadRight(string,LCD_WIDTH," "));
   for (i = 0; i < padded_len - len; i++) {
     if (len - i > 0) string[padded_len-len+i] = string[i];
     string[i] = pad;
   }   
   return string;
}


char* lineprint(char *text, char* str, int position) {
   if (strlen(text) > LCD_WIDTH){
      memcpy(str, &text[position], LCD_WIDTH);
   }else{
      strcpy(str,text);
   }   
   if (strlen(str) < LCD_WIDTH) strcpy(str,StringPadRight(str,LCD_WIDTH," "));
   //str[LCD_WIDTH] = '\n';
   return str;
}

